const fs = require('fs');
const path = require('path');

const CONFIG = require('../Constants/Config');
const Invoice = require('../Model/Invoice');
const Task = require("../Model/Task");
const { workType } = require('../Types/Task');

const TaskController = require('./Task.controller');
const UtilityController = require('./Utilities.controller');

const BranchServices = require('../Services/Branch.services');
const CaseServices = require('../Services/Case.services');
const InvoiceServices = require('../Services/Invoice.services');
const nInvoiceServices = require('../Services/CreateServices')('nInvoice');

/**
 * @description This is used for get list of all invoices
 * @param {*} req 
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, invoices: Array }
 * - { status : false, error : string }
 */
module.exports.getAllInvoices = async (req, res) => {
    try {
        let query = {};
        var count = 0;

        const InvoiceResult = await InvoiceServices.getAllInvoices(query, {}, req.local.limit, req.local.startIndex);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);

        return res.status(200).send({ status: true, invoices: InvoiceResult.invoices, count: count });
    }
    catch (e) {
        console.log("🚀 ~ file: Invoice.controller.js:48 ~ module.exports.getAllInvoices= ~ e:", e)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for update invoice
 * @param {*} req.params.id from that get invoiceId and from req.local.invoice get invoice Object
 * @param {*} res 
 * @returns Object invoice_res { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.deleteSubInvoice = async (req, res) => {
    try {
        const id = req.params.id;
        const subId = req.params.subId;
        console.log("🚀 ~ file: Invoice.controller.js:103 ~ module.exports.deleteSubInvoice ~ subId:", id, subId)

        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const Invoice_res = await InvoiceServices.updateInvoice(id, {
            $pull: {
                "invoices": {
                    "_id": subId
                }
            }
        });
        if (!Invoice_res.status) return res.status(500).send(Invoice_res);
        let invoiceResult = Invoice_res.invoices.toJSON();

        return res.status(200).send({ status: true, invoices: invoiceResult });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};


/**
 * @description This is used for get one invoice
 * @param {*} req.params.id from that get invoiceId  
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, InvoiceResult, Empanelments: Object.values(arr) }
 * - { status : false, error : string }
 */
module.exports.getInvoice = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const InvoiceResult = await InvoiceServices.getInvoice(id);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);

        return res.status(200).send({ status: true, InvoiceResult });
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for create new invoice
 * @param {*} req.local.invoice from that get invoice Object
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.createInvoiceRequest = async (req, res) => {
    try {
        const user = req.local.user;
        const data = req.local.invoice;
        data.organization = { name: user.organization.name, id: user.organization._id };

        const CaseResult = await CaseServices.findCase(data.caseId);
        if (!CaseResult.status || CaseResult.error) return res.status(500).send(CaseResult);
        if (!CaseResult.cases) { return res.status(404).send({ status: false, error: "Case Not Found!" }) }

        const invoice = {
            note: data.note || "SYSTEM GENERATE WHEN Lead/CASE CREATED",
            invoiceType: data.invoiceType || "PROFORMA",
            isAdvance: true,
            amount: data.invoiceAmount || 0,
            requestedBy: user
        };

        const hasInvoiceResult = await InvoiceServices.getAllInvoices({ 'case._id': data.caseId });
        if (!hasInvoiceResult.status || hasInvoiceResult.error) return res.status(500).send(hasInvoiceResult);

        const caseData = CaseResult.cases;
        const advanceInvoiceTaskRes = await TaskController.createTask({ organizationId: caseData.organization._id, caseId: data.caseId, workType: workType.PREPARE_INVOICE });

        if (hasInvoiceResult.invoices.length > 0) {
            const InvoiceResult = await InvoiceServices.updateInvoice(hasInvoiceResult.invoices[0]._id, { $push: { invoices: invoice } });
            return res.status(200).send(InvoiceResult);
        } else {
            const inv = {
                case: caseData,
                organization: caseData.organization,
                gstNo: caseData.gstNo || "",
                invoices: [invoice]
            };
            const InvoiceResult = await InvoiceServices.createInvoice(inv);
            if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);
            return res.status(200).send(InvoiceResult);
        }
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for create new invoice request from attach report/report generate
 * @param {*} data from that get invoice Object
 * @returns Object InvoiceResult { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.createInvoiceRequestSystem = async (data) => {
    try {
        const user = data.user;
        data.organization = { name: user.organization.name, id: user.organization._id };

        const CaseResult = await CaseServices.findCase(data.caseId);
        if (!CaseResult.status || CaseResult.error) return res.status(500).send(CaseResult);
        if (!CaseResult.cases) { return res.status(404).send({ status: false, error: "Case Not Found!" }) }
        const caseData = CaseResult.cases;

        const hasInvoiceResult = await InvoiceServices.getAllInvoices({ 'case._id': data.caseId, invoiceStatus: "OPEN" });
        if (!hasInvoiceResult.status || hasInvoiceResult.error) return res.status(500).send(hasInvoiceResult);

        const existingInvoice = hasInvoiceResult.invoices.length > 0

        if (!existingInvoice) {
            const createInvoiceResult = await nInvoiceServices.create({
                cases: { ...caseData, fees: (caseData.fees.quotedAmount || 0) },
                organization: caseData.organization,
                gstNo: caseData.gstNo || "",
                note: "SYSTEM GENERATE WHEN CHECKER Checked CREATED",
                invoiceType: "PROFORMA",
                amount: (caseData.fees && caseData.fees.quotedAmount) ? caseData.fees.quotedAmount : 0,
                requestedBy: user,
                requestedAt: new Date(),
            });

            if (!createInvoiceResult.status || createInvoiceResult.error) return res.status(500).send(createInvoiceResult);
            return createInvoiceResult;
        } else {
            return { status: false, error: "Open invoices found." };
        }
    }
    catch (err) {
        logger.error(`🚀 ~ file: Invoice.controller.js:189 ~ module.exports.createInvoiceRequestSystem= ~ err: ${JSON.stringify(err)}`)
        return { status: false, error: err };
    }
};

/**
 * @description This is used for update invoice
 * @param {*} req.params.id from that get invoiceId and from req.local.invoice get invoice Object
 * @param {*} res 
 * @returns Object invoice_res { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.updateInvoice = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        let data = req.local.invoice;
        data.totalAmount = parseFloat(data.basic) + parseFloat(data.ops) + parseFloat(data.gst);

        const Invoice_res = await InvoiceServices.updateInvoice(id, data);
        if (!Invoice_res.status) return res.status(500).send(Invoice_res);
        var invoiceResult = Invoice_res.invoices.toJSON();

        return res.status(200).send({ status: true, invoices: invoiceResult });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for update invoice
 * @param {*} req.params.id from that get invoiceId and from req.local.invoice get invoice Object
 * @param {*} res 
 * @returns Object invoice_res { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.invoiceUpload = async (req, res) => {
    try {
        const id = req.params.id; // main invoice id
        const { type } = req.query;
        const invoiceBody = req.local.invoice;

        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const InvoiceResult = await InvoiceServices.getInvoice(id);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);

        let subInvoices = InvoiceResult.invoices.toJSON().invoices || [];
        const caseId = InvoiceResult.invoices.case._id.toString();

        if (!req.file) {
            return res.status(500).send({ status: false, error: "No Invoice File Found" });
        }

        if (!fs.existsSync(process.env.CDN_PATH + '/cases/' + caseId + '/invoices/')) {
            fs.mkdirSync(process.env.CDN_PATH + '/cases/' + caseId + '/invoices/', { recursive: true });
        }
        const fileName = "inv_" + req.file.filename;
        const destination = path.join(process.env.CDN_PATH, "cases", caseId, 'invoices', fileName);

        // call function for move file
        let currentPath = process.cwd() + "/" + req.file.path;
        await UtilityController.moveFile(currentPath, destination);
        const filepath = path.join("cases", caseId, 'invoices', fileName);

        let newData = {
            status: "COMPLETED",
            amount: invoiceBody.invoicedAmount,
            invoiceNo: invoiceBody.invoiceNo,
            completedAt: new Date(),
            invoiceFile: {
                ...req.file,
                fileName,
                path: filepath,
                documentType: { key: "INVOICE", value: "INVOICE" },
                uploadedBy: req.local.user,
                uploadedAt: new Date()
            }
        };

        if (invoiceBody.voucher) {
            newData = { ...newData, voucher: invoiceBody.voucher }
        }

        if (type == "CANCEL") {
            subInvoices = subInvoices.map((inv, id) => {
                if (inv._id == invoiceBody.subInvoiceId) {
                    newData = { ...inv, ...newData, _id: new Date() };
                    inv.status = "CANCELLED";
                    inv.deletedAt = new Date();
                }
                return inv
            });
            subInvoices.push(newData);
        } else {
            subInvoices = subInvoices.map((inv, id) => {
                return (inv._id == invoiceBody.subInvoiceId) ? { ...inv, ...newData } : inv
            });
        }

        const invoicedAmount = subInvoices.reduce((prev, inv) => {
            return prev + ((inv.status === "COMPLETED" && inv.invoiceType == "PROFORMA") ? (+inv.amount) : 0)
        }, 0);

        const updateInvoiceRes = await InvoiceServices.findUpdateInvoice({ _id: id }, { $set: { invoicedAmount: invoicedAmount, "invoices": subInvoices, ...(type == "COMPLETED" && { invoiceStatus: "COMPLETED" }) } });
        if (!updateInvoiceRes.status) return res.status(500).send(updateInvoiceRes);

        return res.status(200).send({ status: true, invoices: updateInvoiceRes.invoices });
    }
    catch (err) {
        console.error("🚀 ~ file: Invoice.controller.js:273 ~ module.exports.invoiceUpload= ~ err:", err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @param {*} req 
 * @param {*} res 
 * @returns Object { status: true, message: "Invoice Deleted" }
 */
module.exports.deleteInvoice = async (req, res) => {
    try {
        const invoiceId = req.params.id;

        const branch_res = await BranchServices.getAllBranches({ 'invoice._id': invoiceId });
        if (!branch_res.status) return res.status(500).send(branch_res);

        if (branch_res.branches.length == 0) {

            const invoice = await InvoiceServices.deleteInvoice(invoiceId);
            if (!invoice.status) return res.status(500).send(invoice);

            return res.status(200).send({ status: true, message: "Invoice Deleted" });
        }
        return res.status(500).send({ status: false, error: "Invoice can't Delete. Because invoice assigned in Branch " });
    }
    catch (e) {
        console.log(e);
        return res.status(500).send({ status: false, error: e });
    }
}

module.exports.getMultipleInvoices = async (req, res) => {
    try {
        const InvoiceResult = await Invoice.aggregate([
            {
                $unwind: { path: "$invoices" }
            },
        ]);

        const invs = InvoiceResult || [];

        const nInvs = await Promise.all(invs.map((inv, id) => {
            delete inv._id;
            let newInv = {
                ...inv,
                cases: [inv.case],
                subId: inv.invoices._id,
                invoiceStatus: inv.invoices.status
            }
            delete inv.invoices._id;
            delete inv.invoices.status;
            newInv = {
                ...newInv,
                ...inv.invoices,
                client: inv.case.client || {},
                branch: inv.case.branch || {},
                clientType: (inv.case.client && inv.case.client.clientType) ? inv.case.client.clientType : { label: 'BANK', value: "BANK" }
            };

            return nInvoiceServices.create(newInv);
        })).then((re) => re);

        return res.status(200).send({ status: true, invoices: nInvs });
    }
    catch (e) {
        console.log("🚀 ~ file: Invoice.controller.js:48 ~ module.exports.getAllInvoices= ~ e:", e)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.oldDeleteDuplicateInvoice = async (req, res) => {
    try {
        const InvoiceResult = await Invoice.aggregate([
            {
                $match: {
                    'invoices.1': { $exists: true }
                },
            },
            {
                '$unwind': {
                    "path": "$invoices"
                }
            },
            {
                "$group": {
                    _id: { case: "$case._id", invoiceNo: "$invoices.invoiceNo", amount: "$invoices.amount" },
                    duplicates: {
                        $addToSet: "$$ROOT"
                    }
                }
            }
        ]);

        const invs = InvoiceResult || [];

        const nInvs = await Promise.all(invs.map((inv, id) => {
            let deleteSubInvs = [];
            let hasOne = false;

            inv.duplicates.reverse().map(async (dupInv, id) => {
                // console.log("🚀 ~ file: Invoice.controller.js:106 ~ inv.duplicates.map ~ dupInv:", dupInv)
                if (!hasOne && dupInv.invoices.status == "COMPLETED") {
                    hasOne = true;
                }
                else if (hasOne && dupInv.invoices.status == "COMPLETED") {
                    console.log("will delete: ", dupInv._id, dupInv.invoices._id, dupInv.invoices.invoiceNo);
                    // await nInvoiceServices.deleteById(dupInv._id)
                    const Invoice_res = await InvoiceServices.updateInvoice(dupInv._id, {
                        $pull: {
                            "invoices": {
                                "_id": dupInv.invoices._id
                            }
                        }
                    });
                    return Invoice_res;
                }

            });
        })).then((re) => re);

        return res.status(200).send({ status: true, invoices: nInvs });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for update invoice
 * @param {*} req.params.id from that get invoiceId and from req.local.invoice get invoice Object
 * @param {*} res 
 * @returns Object invoice_res { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.deleteDuplicateInvoice = async (req, res) => {
    try {

        const InvoiceResult = await nInvoiceServices.aggregate([
            {
                $match: {
                    invoiceStatus: "COMPLETED"
                },
            },
            {
                '$sort': {
                    createdAt: -1
                }
            },
            {
                $group: {
                    _id: { invoiceNo: "$invoiceNo" },
                    dups: { $addToSet: "$$ROOT" },
                },
            },
            {
                $match: { "dups.1": { $exists: true } }
            },
        ]);

        const invs = InvoiceResult.data || [];

        const nInvs = await Promise.all(invs.map((inv, id) => {
            return inv.dups.map(async (dupInv, id) => {
                if (id > 0) {
                    console.log("will delete: ", dupInv._id, dupInv.invoiceNo);
                    // await nInvoiceServices.deleteById(dupInv._id)
                }
            });
        })).then((re) => re);

        return res.status(200).send({ status: true, invoices: nInvs });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};
