// const { data } = require('pdfkit/js/reference');
const OrganizationServices = require('../Services/Organization.services');
const PackageServices = require('../Services/Package.services');
const SubscriptionServices = require('../Services/Subscription.services');

const UserServices = require('../Services/User.services');
const RoleServices = require('../Services/Role.services');
const crypto = require('crypto');
const path = require('path');
const fs = require('fs');
const _ = require('loadsh');
const moment = require('moment');
const MailController = require('../Controllers/Mail.controller');
const { PermissionToBinary } = require('../Services/PermissionToBinaryServices');

/**
 * @description  This is used for get list of all organizations and branches of organization
 * @param {*} req.body.type get type i.e. branch or Organization 
 * @param {*} res 
 * @returns Object OrganizationResult  { status: true, Organizations: Array }
 * - { status: false, error: string }
 */
module.exports.getAllOrganizations = async (req, res) => {
    try {
        const type = req.query.type;

        const OrganizationResult = await OrganizationServices.findAll({ type }, {});
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);
        return res.status(200).send(OrganizationResult);
    }
    catch (e) {
        console.log("🚀 ~ file: Organization.controller.js:30 ~ module.exports.getAllOrganizations= ~ e:", e)
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description  This is used for get One organization 
 * @param {*} req.params.id from that get OrganizationId
 * @param {*} res 
 * @returns Object OrganizationResult  { status: true, Organizations: Object }
 * - { status: false, error: string }
 */
module.exports.getOrganization = async (req, res) => {
    try {
        // console.log(req.query)
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const OrganizationResult = await OrganizationServices.getById(id);
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);

        const BranchResult = await OrganizationServices.findAll({ 'organization._id': id, type: 'BRANCH' }, { sort: { createdAt: 1 } });
        if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

        return res.status(200).send({ ...OrganizationResult, branches: BranchResult.Organizations });
    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description  This is used for Create organization 
 * @param {*} req.local.organization from that get Organization Object
 * @param {*} res 
 * @returns Object OrganizationResult  { status: true, organization: string, organizationBranch: Object, user_id: String, message: "Organization Created." }
 * - { status: false, error: string }
 */
module.exports.createOrganization = async (req, res) => {
    try {
        let orgData = req.local.organization;
        const packageD = orgData.package;
        delete orgData.package;

        const hasOrganization = await OrganizationServices.findAll({ name: orgData.name }, {})
        if (!hasOrganization.status || hasOrganization.error) return res.status(500).send(hasOrganization);

        if (hasOrganization.Organizations.length != 0)
            return res.status(500).send({ status: false, error: "Organization with same name Already Exist!" });

        if (req.files.logo) { orgData.logo = req.files.logo[0].path; }
        let documentArr = [];
        if (req.files.documents) {
            await req.files.documents.map(async (img) => {
                documentArr.push({ ...img, path: img.path, });
            })
            orgData.documents = documentArr;
        }

        // CREATE Organization
        orgData.type = "ORGANIZATION";
        const organization_res = await OrganizationServices.createOrganization(orgData);
        if (!organization_res.status)
            return res.status(500).send(organization_res);

        let subscriptions = {};
        let activeSubscription = {};
        if (!!packageD) {
            const PackageResult = await PackageServices.getPackage(packageD);
            console.log("🚀 ~ file: Organization.controller.js:101 ~ module.exports.createOrganization= ~ PackageResult:", PackageResult)
            if (!PackageResult.status || PackageResult.error) return res.status(500).send(PackageResult);
            const package = PackageResult.packages;

            // create SUBSCRIPTION
            const startDate = new Date();
            const endDate = new Date(moment(startDate).add(+package.duration, 'days'));
            const subRes = await SubscriptionServices.create({ organization: organization_res.Organizations, package: PackageResult.packages, startDate, endDate })
            console.log("🚀 ~ file: Organization.controller.js:110 ~ module.exports.createOrganization= ~ subRes:", subRes)

            subscriptions = [subRes.subscriptions];
            activeSubscription = subRes.subscriptions;
        }

        //create USER
        const password = crypto.randomBytes(8).toString('hex');
        const hashPassword = await UserServices.hashPassword(password);
        if (!hashPassword) return res.status(500).send({ status: false, error: 'Internal Server Error!' });

        // get ADMIN role permissions
        const permissionResult = await RoleServices.findByRoleName({ name: 'ADMIN' });
        if (!permissionResult.status) return res.status(401).send({ ...permissionResult });

        const owners = await Promise.all(orgData?.owners?.map(async (usr, id) => {
            const userData = {
                firstName: usr.firstName,
                lastName: usr.lastName,
                email: usr.email,
                mobileNo: usr.mobileNo,
                organization: organization_res.Organizations,
                type: "ORG_ADMIN",
                role: "ADMIN",
                password: hashPassword.hashPassword,
            };

            userData.permission = permissionResult.roles.permission;
            const user = await UserServices.createUser(userData);
            if (!user.status) return res.status(500).send(user);

            // Send Password in Mail to owner emails id
            const sendMail = await this.sendMail(usr.email, password, req.body.link);

            // For Store password in the folder
            const filename = userData.email + '_' + organizationBranch_res.Organizations._id + '.txt';
            const filePath = path.join('data', filename);
            await fs.writeFile(filePath, "email:  " + userData.email + "\npassword: " + password + "\n", function (err) { if (err) throw err; });

            const newUser = user.users.toJSON();
            return newUser;
        })).then(data => { return data }).catch((err) => { console.log("error in owner creation:", err); return {} });

        //update organization
        const updatedOrganization = await OrganizationServices.updateOrganization(organization_res.Organizations._id, { organization: organization_res.Organizations, owners: owners, subscriptions, activeSubscription })
        console.log("🚀 ~ file: Organization.controller.js:156 ~ module.exports.createOrganization= ~ updatedOrganization:", updatedOrganization)
        if (!updatedOrganization.status)
            return res.status(500).send(updatedOrganization);

        // //update branch
        // const updatedOrgBranch = await OrganizationServices.updateOrganization(organizationBranch_res.Organizations._id, { ...orgBranchData, owners: owners })
        // console.log("🚀 ~ file: Organization.controller.js:162 ~ module.exports.createOrganization= ~ updatedOrgBranch:", updatedOrgBranch)
        // if (!updatedOrgBranch.status)
        //     return res.status(500).send(updatedOrgBranch);

        return res.status(200).send({ status: true, organization: organization_res.Organizations._id, message: "Organization Created." })
    }
    catch (err) {
        console.log("🚀 ~ file: Organization.controller.js:148 ~ module.exports.createOrganization= ~ err:", err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description  This is used for update organization 
 * @param {*} req.params.id from that get OrganizationId an from req.local.oragnization get organization object
 * @param {*} res 
 * @returns Object organization_res  { status: true, Organizations: Object }
 * - { status: false, error: string }
 */
module.exports.updateOrganization = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const organization_data = req.local.organization;

        if (req.file) organization_data.logo = req.file.path;

        const organization_res = await OrganizationServices.updateOrganization(id, organization_data);
        if (!organization_res.status) return res.status(500).send(organization_res);

        return res.status(200).send(organization_res);
    }
    catch (err) {
        return res.status(500).send({ status: false, error: err });
    }
};

module.exports.deleteOrganization = async (req, res) => {
    // return res.status(200).send({ message: "Hi There Its  Admin" });
};

module.exports.enableOrganizationStatus = async (req, res) => {
    try {
        const id = req.params.id;
        // console.log("org Id :" + id)
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const organizationData = { status: true };

        const organization_res = await OrganizationServices.updateOrganization(id, organizationData);
        if (!organization_res.status) return res.status(500).send(organization_res);

        return res.status(200).send(organization_res);
    }
    catch (err) {
        return res.status(500).send({ status: false, error: err });
    }
};

module.exports.disableOrganizationStatus = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const organizationData = { status: false };

        const organization_res = await OrganizationServices.updateOrganization(id, organizationData);

        // console.log("Organization : " + organization_res);
        if (!organization_res) return res.status(500).send(organization_res);
        if (!organization_res.status) return res.status(500).send(organization_res);

        return res.status(200).send(organization_res);
    }
    catch (err) {
        return res.status(500).send({ status: false, error: err });
    }
};

module.exports.getOrganizationUsers = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const OrganizationResult = await UserServices.find({ query: { organization: { _id: id } } });
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);

        return res.status(200).send(OrganizationResult);
    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.getOrganizationUser = async (req, res) => {
    try {
        const id = req.params.id;
        const user_id = req.params.user_id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const OrganizationResult = await UserServices.getOrganizationUser({ organization: { _id: id, user_id: user_id } });
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);

        return res.status(200).send(OrganizationResult);
    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.updateOwnerOrgBranch = async (req, res) => {
    try {
        const organizationId = req.params.id;
        const own = req.body;
        if (!(organizationId && organizationId.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const OrganizationResult = await OrganizationServices.getById(organizationId);
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);

        let userData = req.local.users;
        let owners = [];

        const user = await UserServices.find({ query: { email: own.email } });
        if (user.status && user.data.length > 0) return res.status(500).send({ ...user, message: "Email is already registered." });

        if (own._id) {
            const user = await UserServices.findById(own._id);
            if (!user.status) return res.status(500).send(user);
            owners.push(user)
        }
        else {
            //create USER
            const password = crypto.randomBytes(8).toString('hex');
            const hashPassword = await UserServices.hashPassword(password);
            if (!hashPassword) return res.status(500).send({ status: false, error: 'Internal Server Error!' });

            // get ADMIN role permissions
            const permissionResult = await RoleServices.findByRoleName({ name: 'ADMIN' });
            if (!permissionResult.status) return res.status(401).send({ ...permissionResult });

            userData = {
                firstName: own.firstName,
                lastName: own.lastName,
                email: own.email,
                mobileNo: own.mobileNo,
                organization: OrganizationResult.Organizations,
                type: OrganizationResult.Organizations.type == "ORGANIZATION" ? "ORG_ADMIN" : "BRANCH_ADMIN",
                role: "ADMIN",
                password: hashPassword.hashPassword,
            };

            userData.permission = permissionResult.roles.permission;
            const user = await UserServices.createUser(userData);
            if (!user.status) return res.status(500).send(user);

            // Send Password in Mail to owner emails id
            const sendMail = await this.sendMail(own.email, password, req.body.link);

            // For Store password in the folder
            const filename = userData.email + '_' + OrganizationResult.Organizations._id + '.txt';
            const filePath = path.join('data', filename);
            await fs.writeFile(filePath, "email:  " + userData.email + "\npassword: " + password + "\n", function (err) { if (err) throw err; });

            owners.push(user)
        }

        //update organization
        // const updatedOrganization = await OrganizationServices.updateOrganization(organizationId, {  owners: newOwners, })
        const updatedOrganization = await OrganizationServices.updateOrganization(organizationId, {})
        console.log("🚀 ~ file: Organization.controller.js:156 ~ module.exports.createOrganization= ~ updatedOrganization:", updatedOrganization)
        if (!updatedOrganization.status)
            return res.status(500).send(updatedOrganization);

        return res.status(200).send({ ...updatedOrganization, owners });
    }
    catch (e) {
        console.log("🚀 ~ file: Organization.controller.js:299 ~ module.exports.createUser= ~ e:", e)
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.deleteOwnerOrgBranch = async (req, res) => {
    try {
        const organizationId = req.params.id;
        const ownerId = req.params.ownerId;

        if (!(organizationId && organizationId.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const OrganizationResult = await OrganizationServices.getById(organizationId);
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);

        const owners = OrganizationResult.Organizations.owners.filter((val, id) => { return val._id !== ownerId })

        //update organization
        const updatedOrganization = await OrganizationServices.updateOrganization(organizationId, { organization: organization_res.Organizations, owners: owners, })
        console.log("🚀 ~ file: Organization.controller.js:156 ~ module.exports.createOrganization= ~ updatedOrganization:", updatedOrganization)
        if (!updatedOrganization.status)
            return res.status(500).send(updatedOrganization);

        return res.status(200).send(user);
    }
    catch (e) {
        console.log("🚀 ~ file: Organization.controller.js:299 ~ module.exports.createUser= ~ e:", e)
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
}

/**
 * @description This is used for  create organization's Branch 
 * @param {*} req.local.organization from that get organization Object
 * @param {*} res 
 * @returns Object { status: true, organization: string, user_id: String, message  "Organization's Branch Created." }
 * - { status: false, error: "Already Created!" }
 * - { status: false, error: string }
 */
module.exports.createOrganizationBranch = async (req, res) => {
    // TODO - update logo and documents
    // TODO - check owner email in system or not
    try {
        const organizationId = req.params.orgId;
        let orgBranchData = req.local.branch;
        orgBranchData.type = 'BRANCH';

        // find organization
        const OrganizationResult = await OrganizationServices.getById(organizationId);
        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);

        // check branch with same name has or not
        const hasOrgBranch = await OrganizationServices.findAll({ name: orgBranchData.name }, {})
        if (!hasOrgBranch.status || hasOrgBranch.error) return res.status(500).send(hasOrgBranch);
        if (hasOrgBranch.Organizations.length != 0)
            return res.status(500).send({ status: false, error: "Organization with same name Already Exist!" });

        if (req.files.logo) { orgBranchData.logo = req.files.logo[0].path; }
        let documentArr = [];
        if (req.files.documents) {
            await req.files.documents.map(async (img) => {
                documentArr.push({ ...img, path: img.path, });
            })
            orgBranchData.documents = documentArr;
        }
        orgBranchData.organization = OrganizationResult.Organizations;
        console.log("🚀 ~ file: Organization.controller.js:374 ~ module.exports.createOrganizationBranch= ~ orgBranchData:", orgBranchData)

        const orgBranchRes = await OrganizationServices.createOrganization(orgBranchData);
        if (!orgBranchRes.status) return res.status(500).send(orgBranchRes);

        const OrgBranchID = orgBranchRes.Organizations._id;

        //create USER
        const password = crypto.randomBytes(8).toString('hex');
        const hashPassword = await UserServices.hashPassword(password);
        if (!hashPassword) return res.status(500).send({ status: false, error: 'Internal Server Error!' });

        // get ADMIN role permissions
        const permissionResult = await RoleServices.findByRoleName({ name: 'ADMIN' });
        if (!permissionResult.status) return res.status(401).send({ ...permissionResult });

        const owners = await Promise.all(orgBranchData?.owners?.map(async (usr, id) => {
            const userData = {
                firstName: usr.firstName,
                lastName: usr.lastName,
                email: usr.email,
                mobileNo: usr.mobileNo,
                organization: orgBranchRes.Organizations,
                type: "ORG_ADMIN",
                role: "ADMIN",
                password: hashPassword.hashPassword,
            };

            userData.permission = permissionResult.roles.permission;
            const user = await UserServices.createUser(userData);
            if (!user.status) return res.status(500).send(user);

            // Send Password in Mail to owner emails id
            const sendMail = await this.sendMail(usr.email, password, req.body.link);

            // For Store password in the folder
            const filename = userData.email + '_' + OrgBranchID + '.txt';
            const filePath = path.join('data', filename);
            await fs.writeFile(filePath, "email:  " + userData.email + "\npassword: " + password + "\n", function (err) { if (err) throw err; });

            return { ...user.users.toJSON(), id: user.users._id };
        })).then(data => { return data }).catch((err) => { console.log("error in owner creation:", err); return {} });

        // // create subscription
        // if (!!orgBranchData.package) {
        //     const PackageResult = await PackageServices.getPackage(orgBranchData.package);
        //     if (!PackageResult.status || PackageResult.error) return res.status(500).send(PackageResult);
        //     const package = PackageResult.packages;

        //     const startDate = new Date();
        //     const endDate = new Date(moment(startDate).add(+package.duration, 'days'));
        //     const subRes = await SubscriptionServices.create({ organization: orgBranchRes.Organizations, package: PackageResult.packages, startDate, endDate })

        //     orgBranchData.subscriptions = [subRes.subscriptions];
        //     orgBranchData.activeSubscription = subRes.subscriptions;
        // }

        orgBranchData = {};
        orgBranchData.owners = owners;

        //update branch
        const updatedOrgBranch = await OrganizationServices.updateOrganization(OrgBranchID, orgBranchData)
        console.log("🚀 ~ file: Organization.controller.js:162 ~ module.exports.createOrganization= ~ updatedOrgBranch:", updatedOrgBranch)
        if (!updatedOrgBranch.status)
            return res.status(500).send(updatedOrgBranch);

        return res.status(200).send({ status: true, organization: updatedOrgBranch.Organizations, message: "Organization's Branch Created." })

    }
    catch (err) {
        console.log("🚀 ~ file: Organization.controller.js:391 ~ module.exports.createOrganizationBranch= ~ err:", err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for  update organization's Branch 
 * @param {*} req.params.id from that get orgId and req.local.organization from that get organization Object
 * @param {*} res 
 * @returns Object { status: true, organization: Object }
 * - { status: false, error: string }
 */
module.exports.updateOrganizationBranch = async (req, res) => {
    // TODO - update logo and documents
    try {
        const organizationId = req.params.orgId;
        const branchId = req.params.branchId;

        // find organization
        const BranchResult = await OrganizationServices.getById(branchId);
        if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

        const datas = req.local.branch;

        if (req.file) datas.logo = req.file.path;

        // // create subscription
        // if (!!datas.package) {
        //     const PackageResult = await PackageServices.getPackage(datas.package);
        //     if (!PackageResult.status || PackageResult.error) return res.status(500).send(PackageResult);
        //     const package = PackageResult.packages;

        //     const startDate = new Date();
        //     const endDate = new Date(moment(startDate).add(+package.duration, 'days'));
        //     const subRes = await SubscriptionServices.create({ organization: BranchResult.Organizations, package: PackageResult.packages, startDate, endDate })

        //     // datas.subscriptions = [subRes.subscriptions];
        //     datas.activeSubscription = subRes.subscriptions;
        // }

        const organization_res = await OrganizationServices.updateOrganization(branchId, datas);
        if (!organization_res.status) return res.status(500).send(organization_res);

        return res.status(200).send({ status: true, organization: organization_res });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used  for get list of all branches for particular Organization 
 * @param {*} req.params.id from that get orgId 
 * @param {*} res 
 * @returns Object OrganizationResult { status: true, Organizations: Array }
 * - {status : false , error : string }
 */
module.exports.getBranchesOrganizationWise = async (req, res) => {
    try {
        const id = req.params.id;

        const OrganizationResult = await OrganizationServices.findAll({ 'organization._id': id, type: "BRANCH" }, {});

        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);
        return res.status(200).send(OrganizationResult);

    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used  for get list of all branches for particular Organization 
 * @param {*} req.params.id from that get orgId 
 * @param {*} res 
 * @returns Object OrganizationResult { status: true, Organizations: Array }
 * - {status : false , error : string }
 */
module.exports.getBranchesOrganizationWise = async (req, res) => {
    try {
        const id = req.params.id;

        const OrganizationResult = await OrganizationServices.findAll({ 'organization._id': id, type: "BRANCH" }, {});

        if (!OrganizationResult.status || OrganizationResult.error) return res.status(500).send(OrganizationResult);
        return res.status(200).send(OrganizationResult);

    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This Function is used For Send Credential to Mail
 * @param {String} email 
 * @param {String} password 
 * @param {String} link 
 * @returns Object mailResult {status : true, result : String}
 */
module.exports.sendMail = async (email, password, link) => {
    try {
        const html = "<p>Hello</p><p>Greetings of the day</p><p> Email: " + email + " <br>Password: " + password + " </p><p>You can update your password from this link: " + link + " </p>";
        const subject = "Login Credential";
        const from = "RealValue";
        const to = email;
        const message = "";
        const attachments = "";

        const mailResult = await MailController.SendMail(to, subject, message, attachments, html);

        return mailResult;
    }
    catch (err) {
        console.log(err);
        // return res.status(500).send({ status: false, error: err });
    }
}