const ReportTemplateServices = require('../Services/CreateServices')('ReportTemplate');
const ClientServices = require('../Services/Client.services');
const ZoneServices = require('../Services/Zone.services');
const BranchServices = require('../Services/Branch.services');
const OrganizationServices = require('../Services/Organization.services');
const { logger } = require('../logger');

module.exports.getAllTemplateMaster = async (req, res) => {
    try {
        let query = {};
        let count = 0;
        // query['organization._id'] = req.local.user.organization._id;

        const { client, branch, zone, searchValue, propertyType, typeOfReport, status } = req.query;

        if (searchValue) {
            query['$or'] = [
                { status: { $regex: searchValue, $options: 'i' } },
                { propertyType: { $regex: searchValue, $options: 'i' } },
                { "client.name": { $regex: searchValue, $options: 'i' } },
                { "client.initial": { $regex: searchValue, $options: 'i' } },
                { "branch.name": { $regex: searchValue, $options: 'i' } }
            ]
        }
        // For Advanced By ID
        if (branch) { query['branch._id'] = branch; }
        if (client) { query['client._id'] = client; }
        if (zone) { query['zone._id'] = zone; }
        if (propertyType) { query['propertyType'] = { $regex: propertyType, $options: 'i' } }
        if (typeOfReport) { query['typeOfReport'] = { $regex: typeOfReport, $options: 'i' } }
        if (status) { query['status'] = { $regex: status, $options: 'i' } }

        const ReportTemplateResult = await ReportTemplateServices.aggregate([
            {
                '$match': { ...query },
            },
            {
                '$group': {
                    _id: {
                        client: "$client._id",
                        branchId: "$branch._id",
                        zoneId: "$zone._id",
                    },
                    client: { $first: "$client" },
                    branch: { $first: "$branch" },
                    zone: { $first: "$zone" },
                    templateCount: { $sum: 1 },
                    templates: { $addToSet: "$$ROOT" },
                }
            },
            {
                '$sort': {
                    'client.name': 1
                }
            }
        ]);

        count = ReportTemplateResult.data.length || 0;
        const reports = ReportTemplateResult.data.slice(req.local.startIndex, req.local.startIndex + req.local.limit);
        if (!ReportTemplateResult.status || ReportTemplateResult.error) return res.status(500).send(ReportTemplateResult);
        return res.status(200).send({ status: true, reportTemplates: reports, count: count });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getAllTemplateMaster= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

/**
 * @description This is used for get list of all ReportTemplates 
 * @param {*} req 
 * @param {*} res 
 * @returns Object ReportTemplateResult {status : true , reportTemplates : Array }
 * -  {status : false , error : string }
 */
module.exports.getAllReportTemplates = async (req, res) => {
    try {
        let query = {};
        var count = 0;
        // query['organization._id'] = req.local.user.organization._id;

        const { client, branch, zone, searchValue, propertyType, typeOfReport, status } = req.query;

        if (searchValue) {
            query['$or'] = [
                { propertyType: { $regex: searchValue, $options: 'i' } },
                { typeOfReport: { $regex: searchValue, $options: 'i' } },
                { "client.name": { $regex: searchValue, $options: 'i' } },
                { "client.initial": { $regex: searchValue, $options: 'i' } },
                { "branch.name": { $regex: searchValue, $options: 'i' } }
            ]
        }
        // For Advanced By ID
        if (branch) { query['branch._id'] = branch; }
        if (client) { query['client._id'] = client; }
        if (zone) { query['zone._id'] = zone; }
        if (propertyType) { query['propertyType'] = { $regex: propertyType, $options: 'i' } }
        if (typeOfReport) { query['typeOfReport'] = { $regex: typeOfReport, $options: 'i' } }
        if (status) { query['status'] = status }

        const ReportTemplateResultCount = await ReportTemplateServices.count(query);
        if (!ReportTemplateResultCount.status || ReportTemplateResultCount.error) return res.status(500).send(ReportTemplateResultCount);

        count = ReportTemplateResultCount.count;

        const ReportTemplateResult = await ReportTemplateServices.find({ query: query, limit: req.local.limit, start: req.local.startIndex });
        if (!ReportTemplateResult.status || ReportTemplateResult.error) return res.status(500).send(ReportTemplateResult);

        return res.status(200).send({ status: true, reportTemplates: ReportTemplateResult.data, count: count });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getAllReportTemplates= ~ e:`, e)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for get one ReportTemplate
 * @param {*} req.params.id from that get ReportTemplateId 
 * @param {*} res 
 * @returns Object ReportTemplateResult {status : true , reportTemplates : Object }
 * -  {status : false , error : string }
 */
module.exports.getReportTemplate = async (req, res) => {
    try {
        const id = req.params.id;
        // console.log(id)
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const ReportTemplateResult = await ReportTemplateServices.findById(id);
        if (!ReportTemplateResult.status || ReportTemplateResult.error) return res.status(500).send(ReportTemplateResult);

        return res.status(200).send({ status: true, reportTemplates: ReportTemplateResult.data });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getReportTemplate= ~ e:`, e.toString())
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for create new reportTemplate
 * @param {*} req.local.reportTemplate  from that get reportTemplates Object
 * @param {*} res 
 * @returns Object ReportTemplateResult {status : true , reportTemplates : Object }
 * -  {status : false , error : string }
 */
module.exports.createReportTemplate = async (req, res) => {
    try {
        const data = req.local.reportTemplate;

        const alreadyExist = await ReportTemplateServices.find({ query: { 'client._id': data.client, 'zone._id': data.zone, 'branch._id': data.branch, propertyType: data.propertyType, typeOfReport: data.typeOfReport } });
        if (!alreadyExist.status || alreadyExist.error) return res.status(500).send(alreadyExist);

        if (alreadyExist.data.length == 0) {
            const orgId = req.local.user.organization;

            const Organizations = await OrganizationServices.getById(orgId);
            if (!Organizations.status || Organizations.error) return res.status(500).send(Organizations);
            data.organization = Organizations.Organizations;

            if (data.client) {
                const client = await ClientServices.getClientById(data.client);
                if (!client.status) return res.status(500).send(client);

                data.client = client.clients;;
            }
            if (data.zone) {
                const ZoneResult = await ZoneServices.getZone(data.zone);
                if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

                data.zone = ZoneResult.zones;
            }
            if (data.branch) {
                const BranchResult = await BranchServices.getBranch(data.branch);
                if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

                data.branch = BranchResult.branches;
            }
            const ReportTemplateResult = await ReportTemplateServices.create(data);
            if (!ReportTemplateResult.status || ReportTemplateResult.error) return res.status(500).send(ReportTemplateResult);

            return res.status(200).send({ status: true, reportTemplates: ReportTemplateResult.data });
        } else {
            return res.status(500).send({ status: false, error: "Already Created!" });
        }

    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.createReportTemplate= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

/**
 * @description This is used for update ReportTemplate
 * @param {*} req.params.id from that get ReportTemplateId and from req.local.reportTemplate get ReportTemplate Object
 * @param {*} res 
 * @returns Object ReportTemplate_res {status : true , reportTemplates : Object }
 * -  {status : false , error : string }
 */
module.exports.updateReportTemplate = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const datas = req.local.reportTemplate;

        if (datas.client) {
            const client = await ClientServices.getClientById(datas.client);
            if (!client.status) return res.status(500).send(client);

            datas.client = client.clients;;
        }
        if (datas.zone) {
            const ZoneResult = await ZoneServices.getZone(datas.zone);
            if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

            datas.zone = ZoneResult.zones;

        }
        else {
            datas.zone = {}
        }
        if (datas.branch) {
            const BranchResult = await BranchServices.getBranch(datas.branch);
            if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

            datas.branch = BranchResult.branches;
        }
        else {
            datas.branch = {}
        }

        const ReportTemplate_res = await ReportTemplateServices.updateById(id, datas);
        if (!ReportTemplate_res.status) return res.status(500).send(ReportTemplate_res);

        return res.status(200).send({ status: true, reportTemplates: ReportTemplate_res.data });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.updateReportTemplate= ~ err:`, err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

/**
 * @description This is used for  Check template  is already or not  
 * @param {*} req.body from that get Object
 * @param {*} res 
 * @returns Object { status: true }
 * - { status: false, error: "Already Created!" }
 */
module.exports.checkAlreadyExistTemplate = async (req, res) => {
    try {

        const case_data = req.body;
        let query = { 'client._id': case_data.client._id, propertyType: case_data.propertyType, typeOfReport: case_data.typeOfReport };

        if (!!case_data._id) {
            query._id = { $ne: case_data._id };
        }
        if (!!case_data.branch){
            query = { ...query, 'branch._id': case_data.branch };
        }
        // console.log(case_data, " ----------- req ------------ ")
        const alreadyExistInClient = await ReportTemplateServices.find({ query });
        if (!alreadyExistInClient.status || alreadyExistInClient.error) return res.status(500).send(alreadyExistInClient);

        if (alreadyExistInClient.data.length !== 0) {
            return res.status(500).send({ status: false, error: "Template Already Exists!" });
        }
        else {
            return res.status(200).send({ status: true });
        }
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.checkAlreadyExistTemplate= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
};