const Task = require('../Model/Task');

/**
 * @description This is used for get All tasks
 * @param {*} query 
 * @param {*} displayFields 
 * @param {*} startIndex 
 * @param {*} limit 
 * @returns Object { status: true, tasks: Array }
 *  { status: false, error: String }
 */
module.exports.findAll = async (query, { display = [], sort = { createdAt: -1 }, populate = [] } = { display: [], sort: { createdAt: -1 }, populate: [] }, startIndex, limit) => {
    const count = await Task.countDocuments(query);

    return Task.find(query, display).populate(populate).sort(sort).skip(startIndex).limit(limit)
        .then((Tasks) => {
            return { status: true, tasks: Tasks, count };
        }).catch((e) => {
            return { status: false, error: e };
        });
};

/**
 * @description This is used aggregate
 * @param {*} query 
 * @returns Object { status: true, tasks: Array }
 *  { status: false, error: String }
 */
module.exports.aggregate = async (query) => {
    return Task.aggregate(query)
        .then((Tasks) => {
            return { status: true, tasks: Tasks };
        }).catch((e) => {
            return { status: false, error: e };
        });
};

/**
 * @description This is used for get One task from an Id
 * @param {String} id 
 * @returns  Object { status: true, tasks: Object }
 *  { status: false, error: String }
 */
module.exports.findById = async (id, { populate = [] } = {}) => {

    return Task.findById(id).populate(populate).then((Task) => {
        return { status: true, tasks: Task };
    }).catch((e) => {
        return { status: false, error: e };
    });
};

/**
 * @description Thisis used for create Task
 * @param {Object} taskData 
 * @returns Object { status: true, tasks: Object }
 *  { status: false, error: String }
 */
module.exports.create = async (taskData) => {

    return Task.create(taskData)
        .then((result) => {
            return { status: true, tasks: result };
        })
        .catch((err) => {
            return { status: false, error: err }
        })
};

/**
 * @description This is Used for update Task
 * @param {String} id 
 * @param {Object} taskData 
 * @returns Object { status: true, tasks: Object }
 *  { status: false, error: String }
 */
module.exports.update = async (id, taskData) => {
    try {
        const taskResult = await Task.findOneAndUpdate({ _id: id }, taskData, { new: true });
        if (!taskResult) return { status: false, error: "Task not found" };

        return { status: true, tasks: taskResult };
    }
    catch (err) {
        return { status: false, error: err };
    }
};

/**
 * @description This is used for update Many tasks
 * @param {Object} find_query 
 * @param {Object} task_data 
 * @returns Object { status: true, tasks: result }
 * -  { status: false, err }
 */
module.exports.updateByQuery = async (find_query, task_data) => {
    return Task.updateMany(find_query, task_data, { new: true })
        .then((result) => {
            return { status: true, tasks: result }
        })
        .catch((err) => {
            return { status: false, err }
        })
};

/**
 * @description This is used for get All tasks
 * @param {*} query 
 * @param {*} displayFields 
 * @param {*} startIndex 
 * @param {*} limit 
 * @returns Object { status: true, tasks: Array, count: Number }
 *  { status: false, error: String }
 */
module.exports.count = async (query) => {
    return Task.countDocuments(query)
        .then((count) => {
            return { status: true, count };
        }).catch((e) => {
            return { status: false, error: e };
        });
};

/**
 * @description This is used for get All Completed Tasks details  
 * @param {Object} taskData 
 * @param {String} orgId 
 * @returns Object { status: true, tasks: Array }
 * -{ status: false, error: string }
 */
module.exports.nearByLocation = async (taskData, orgId) => {

    return Task.aggregate([
        // { '$match': { taskStatus: "CLOSED" , "organization._id" : orgId} },
        {
            "$geoNear": {
                near: {
                    type: "Point",
                    coordinates: taskData
                },
                maxDistance: 1000,
                spherical: true,
                query: {
                    taskStatus: "COMPLETED",
                    // "organization._id": orgId
                },
                distanceField: "dist"
            }
        },
        {
            $project: {
                taskNumber: 1, "client.name": 1, siteMeasurement: { dateOfConstruction: { year: 1 }, noOfFloors: 1 }, "branch.name": 1, "routeMap.finalLatLng.coordinates": 1, dist: 1, finalValuationAreaString: 1,
            }
        }
    ])
        .then(result => {
            return { status: true, tasks: result }
        }).catch(err => {
            return { status: false, error: err }
        })

};
