const cron = require('node-cron');
const { cronLogger } = require('./logger');
const TaskServices = require('./Services/CreateServices')('Task');
const AttendanceServices = require('./Services/CreateServices')('Attendance');
const AttendanceSummaryServices = require('./Services/CreateServices')('AttendanceSummary');

function ClientVisitAutoComplete() {
    cron.schedule('*/30 * * * *', async () => {
        console.log('Running  ClientVisitAutoComplete cron job...');
        try {
            const twoHoursAgo = new Date(Date.now() - 2 * 60 * 60 * 1000);
            // Find documents where updatedAt is greater than 2 hours ago
            const documentsToUpdate = await TaskServices.find({ query: { workType: "CLIENT_VISIT", status: "MY_TASK", updatedAt: { $lt: twoHoursAgo } }, limit: 1000000 });
            console.log("query: ", { query: { workType: "CLIENT_VISIT", status: "MY_TASK", updatedAt: { $lt: twoHoursAgo } }, limit: 1000000 }, " - ", documentsToUpdate.count);
            if (documentsToUpdate.status) {

                for (const doc of documentsToUpdate.data) {
                    // complete each tasks which is open for more then 2 hr
                    doc.timeLog.push({
                        time: new Date(),
                        type: "END",
                        geo: { type: 'Point', coordinates: [0, 0] }
                    });
                    doc.remarks.push({
                        id: Math.floor(100000 + Math.random() * 900000),
                        remark: 'AUTO COMPETED BY SERVER.',
                        createdAt: new Date(),
                        creditOfficerId: null,
                        firstName: "-",
                        lastName: "-",
                        role: "-",
                    });
                    doc.status = "COMPLETED";
                    await doc.save();
                    console.log(`Updated document with ID ${doc._id}`);
                }
            }
        } catch (error) {
            cronLogger.error('Error in ClientVisitAutoComplete cron job:' + error.toString());
        }
    });
}

async function autoCloseAttendance() {
    cron.schedule('59 23 * * *', async () => {
        console.log('Running  autoCloseAttendance cron job...');
        try {
            console.log('Running attendance auto-close...');

            const currentDate = new Date();
            const dayStart = new Date(currentDate.setHours(0, 0, 0, 0));
            const dayEnd = new Date(currentDate.setHours(23, 59, 59, 999));

            // Fetch open attendance records for the current day
            const openAttendances = await AttendanceServices.Model.find({
                query: {
                    checkIn: { $exists: true },
                    "checkOut.time": { $eq: null },
                    date: { $gte: dayStart, $lt: dayEnd },
                },
                limit: 1000000,
            });

            console.log("Open attendances to process: ", openAttendances.count);

            if (openAttendances.status && openAttendances.data.length > 0) {
                for (const attendance of openAttendances.data) {
                    // Auto-close attendance for the day
                    attendance.checkOut = {
                        time: dayEnd,
                        geo: { type: "Point", coordinates: [0, 0] },
                        note: "AUTO CLOSE"
                    };

                    // Calculate working hours (in hours)
                    attendance.workingHours = attendance.checkIn.time
                        ? Math.max(
                            0,
                            (dayEnd - new Date(attendance.checkIn.time)) / 3600000 // Convert ms to hours
                        )
                        : 0;

                    // Update present status
                    attendance.present = true;

                    // Save the updated attendance record
                    await attendance.save();
                    console.log(`Auto-closed attendance for ID ${attendance._id}`);
                }
            } else {
                console.log("No open attendances to process.");
            }
        } catch (error) {
            cronLogger.error("Error in auto-close attendance: " + error.toString());
        }
    });
}

async function autoCalculateAttendanceSummary() {
    cron.schedule('0 1 * * *', async () => {
        cronLogger.error('Running  autoCalculateAttendanceSummary cron job...');
        try {
            console.log('Running attendance summary calculation...');

            // Calculate start and end time for the previous day
            const now = new Date();
            const previousDayStart = new Date(now.setDate(now.getDate() - 3)).setHours(0, 0, 0, 0);
            const previousDayEnd = new Date(now).setHours(23, 59, 59, 999);

            // Fetch all attendance records for the previous day
            const attendances = await AttendanceSummaryServices.Model.find({
                query: {
                    date: { $gte: new Date(previousDayStart), $lte: new Date(previousDayEnd) },
                },
                limit: 1000000,
            });

            console.log(`Fetched ${attendances.data.length} attendance records for the previous day.`);

            // Group attendances by userId
            const attendanceByUser = {};
            for (const record of attendances.data) {
                const userId = record.userId.toString();
                if (!attendanceByUser[userId]) {
                    attendanceByUser[userId] = [];
                }
                attendanceByUser[userId].push(record);
            }

            // Process each user's attendance summary
            for (const [userId, userAttendances] of Object.entries(attendanceByUser)) {
                let totalWorkingHours = 0;
                let isPresent = false;

                // Calculate total working hours and presence
                for (const attendance of userAttendances) {
                    if (attendance.workingHours) {
                        totalWorkingHours += attendance.workingHours;
                    }
                    if (attendance.present) {
                        isPresent = true;
                    }
                }

                // Determine absent/present status
                const status = isPresent ? 'Present' : 'Absent';

                // Create or update calendar entry for the user
                const calendarEntry = {
                    userId,
                    date: new Date(previousDayStart),
                    totalWorkingHours,
                    status,
                };

                await AttendanceSummaryServices.Model.updateOrCreate(
                    { userId, date: new Date(previousDayStart) }, // Query
                    calendarEntry // Data to update or create
                );

                console.log(`Logged attendance summary for user ${userId}:`, calendarEntry);
            }

            console.log('Attendance summary calculation completed.');
        } catch (error) {
            cronLogger.error('Error in attendance summary calculation:', error);
        }
    });
}

function startCronJobs() {
    ClientVisitAutoComplete();
    autoCloseAttendance();
    autoCalculateAttendanceSummary();
}

module.exports = { startCronJobs };
