const express = require("express");
const fs = require('fs');
const path = require('path');
const moment = require('moment');
const _ = require('loadsh');
const json2xls = require('json2xls');

const Case = require("../Model/Case");
const User = require("../Model/User");
const { logger } = require('../logger');
const isAuth = require("../Middleware/isAuth");
const { sendExpoNotification } = require("../Services/ExpoNotification.services");
const CaseServices = require('../Services/CreateServices')('Case');

const nInvoiceServices = require('../Services/CreateServices')('nInvoice');
const PaymentServices = require('../Services/CreateServices')('Payment');
const TaskServices = require('../Services/Task.services');
const router = express.Router();

router.get('/dev/monthly-report', async (req, res) => {
    try {
        let startDate = req.query.start;
        let endDate = req.query.end;

        if ((!startDate || !endDate || !moment(startDate, "DD-MM-YYYY").isValid() || !moment(endDate, "DD-MM-YYYY").isValid())) {
            let date = new Date();
            endDate = new Date(date.getFullYear(), date.getMonth() + 1, 0);
            startDate = new Date(date.getFullYear(), date.getMonth(), 1);
        }
        else {
            startDate = new Date(moment(startDate, "DD-MM-YYYY").startOf('day'));
            endDate = new Date(moment(endDate, "DD-MM-YYYY").endOf('day'));
        }
        console.log("start: ", startDate, "end: ", endDate)

        const result = await Case.aggregate([
            {
                '$match': { createdAt: { $gte: startDate, $lte: endDate } }
            },
            {
                '$lookup': {
                    from: 'tasks',
                    localField: '_id',
                    foreignField: 'caseId',
                    as: 'tasks'
                }
            },
            {
                '$project': {
                    caseNumber: 1,
                    "client": 1,
                    "branch.name": 1,
                    "address.buildingName.value": 1,
                    fullAddress: 1,
                    ownerName: 1,
                    proposePurchaserName: 1,
                    "propertyType.value": 1,
                    "typeOfReport.label": 1,
                    externalReportDetails: 1,
                    flatRate: 1,
                    valuationDate: 1,
                    reportGeneratedDate: 1,
                    outward: 1,
                    "tasks.workType": 1,
                    "tasks.users": 1,
                    caseReferredBy: 1,
                    staff: 1,
                    caseStatus: 1,
                    ownerName: 1,
                    outward: 1,
                    createdAt: 1,
                    organization: 1,
                    remarks: 1,
                }
            }
        ]).allowDiskUse(true);

        const newResult = await Promise.all(result.map((cs, id) => {
            return {
                CASE_NO: cs.caseNumber,
                organization: cs.organization.shortName,
                BANK: cs?.client?.name || "-",
                BRANCH: cs?.branch?.name || "-",
                Building_Name: cs.address.buildingName.value || "",
                clientName: cs.proposePurchaserName || cs.ownerName || "",
                Property_Type: (cs?.propertyType?.value || "") + " " + (cs?.typeOfReport?.label || ""),
                ReportGenerateDate: moment(cs?.outward?.date).isValid() ? moment(cs?.outward?.date).format('DD-MM-YYYY') : (cs?.valuationDate || cs?.reportGeneratedDate),
                maker: (!!cs.staff?.maker?.firstName) ? (cs.staff.maker.firstName + " " + cs.staff.maker.lastName) : "",
                checker: (!!cs.staff?.checker?.firstName) ? (cs.staff.checker.firstName + " " + cs.staff.checker.lastName) : "",
                rateVerifier: (!!cs.staff?.rateVerifier?.firstName) ? (cs.staff.rateVerifier.firstName + " " + cs.staff.rateVerifier.lastName) : "",
                siteEngineer: (!!cs.staff?.siteEngineer?.firstName) ? (cs.staff.siteEngineer.firstName + " " + cs.staff.siteEngineer.lastName) : "",
                pickupPerson: (!!cs.staff?.pickupPerson?.firstName) ? (cs.staff.pickupPerson.firstName + " " + cs.staff.pickupPerson.lastName) : "",
                deliveryPerson: (!!cs.staff?.deliveryPerson?.firstName) ? (cs.staff?.deliveryPerson.firstName + " " + cs.staff.deliveryPerson.lastName) : "",
                Address: cs.fullAddress || "-",
                outward: cs.outward.number || "-",
                remark: _.last(cs.remarks || [])?.note || "",
                status: cs.caseStatus,
                createdAt: cs.createdAt,
            }
        })).then((resu) => {
            return resu;
        });

        const xls = json2xls(newResult);

        // Write Excel file to a temporary file
        const filePath = __dirname + `/monthly-report.xlsx`;
        fs.writeFileSync(filePath, xls, 'binary');

        return res.sendFile(filePath, {}, (err) => {
            if (err) {
                logger.error('Error occurred while sending file:', err);
                return res.status(err.status).end();
            } else {
                fs.unlinkSync(filePath); // Remove the temporary file after sending
            }
        });
    }
    catch (e) {
        logger.error("dev routers dev/monthly-report: " + e.toString());
        return res.status(500).send({ status: false, error: e.toString() });
    }
});

router.get('/dev/devQs', isAuth, async (req, res) => {
    return fs.readFile(path.join(__dirname + '/../public/devQs.json'), (error, data) => {
        if (error) {
            console.warn("Error in Configuration file read:", error);
            return res.status(500).send({ status: false, error: JSON.stringify(e) })
        };
        console.log("=> ", JSON.parse(data));
        return res.status(200).send({ status: true, devQs: JSON.parse(data) });
    });

});

router.post('/dev/devQs', isAuth, async (req, res) => {
    const qu = { skip: 0, limit: 10, populate: [], project: {}, ...req.body };
    const models = { 'case': Case, 'Case': Case };

    try {
        if (qu.queryType == "aggregate") {
            const result = await models[qu.model][qu.queryType](qu.query);
            return res.status(200).send({ result })
        }
        else if (qu.queryType == "find") {
            const result = await models[qu.model][qu.queryType](qu.query, qu.project).skip(qu.skip).limit(qu.limit).populate(qu.populate).sort(qu.sort);
            return res.status(200).send({ result })
        }
        else {
            return res.status(404).send({ error: "Sorry bro you are looking at wrong place." })
        }
    }
    catch (e) {
        console.log("🚀 ~ file: Dev.routes.js:47 ~ router.post ~ e:", e)
        return res.status(500).send({ status: false, error: JSON.stringify(e) })
    }
});

router.get('/check-notification', async (req, res) => {
    let tokens = await User.find({ notificationToken: { $exists: 1 } }, { notificationToken: 1 });

    tokens = tokens.reduce((prev, curr) => [...prev, curr.notificationToken], []);

    const sent = await sendExpoNotification(tokens, "Testing Notification", {});
    return res.status(200).send({ status: true, sent })
})

router.get('/igr-masters/documents', async (req, res) => {
    return res.status(200).send({
        status: true,
        data: ['सेल सर्टिफिकेट', 'सर्टिफिकेट ऑफ सेल', 'विक्री दाखला', 'विक्री प्रमाणपत्र', 'फरोक्तखरेदीखत', 'असाईनमेंट डीड', 'खुषखरेदीखत', 'विक्रीपत्र', 'ट्रान्सफर डीड', 'हस्तांतरणपत्र', 'कायम खुषखरेदीखत', 'तबदीलीचा करारनामा', 'अँग्रीमेंट टू असाईनमेंट', 'इकरारनामा', 'करारनामा', 'खरेदीखत', 'विक्री करारनामा', 'अँग्रीमेंट टू सेल ऑफ ऑफिस', 'अँग्रीमेंट टू सेल ऑफ शॉप', 'सेल डीड', 'कन्व्हेन्स डीड', 'अविभाज्य हिश्याची पूर्ण विक्री', 'अविभाज्य हिश्याची अंशतः विक्री', 'अँग्रीमेंट टू सेल ऑफ फ्लॅट', 'अभिहस्तांतरणपत्र', 'साठेखत', 'सप्लीमेंट्री अँग्रीमेंट', 'पुरवणी करारनामा', 'अँग्रीमेंट टू सेल', 'जादा बोजाचे गहाणखत', 'भाडेपट्टा', 'लीजडीड', 'लिव्ह अँणड लायसन्स', 'भाडेकरार', 'किरायानामा', 'अँग्रीमेंट टू लीज', 'डीड ऑफ मॉरगेज', 'गहाणखत', 'मॉरगेज डीड', 'रिलीज डीड', 'हक्कसोडपत्र', '59 - हस्तांतरण', '36 - अ - लिव्ह अॅड लायसन्सेस', '36 - अ - Leave and Licenses']
    });
});


router.get('/dev/print_task', async (req, res) => {
    let taskD = {
        "geo": {
            "type": "Point",
            "coordinates": [0, 0]
        },
        "workType": "PRINT",
        "duration": 0,
        "priority": 0,
        "status": "OPEN",
        "users": [],
        "timer": [],
        "review": []
    };

    const caseD = await Case.find({ caseNumber: "" + req.query.cn });
    if (!caseD && !caseD._id) return res.status(500).send(caseD);

    let c = caseD[0];
    taskD.caseId = c._id;
    taskD.organizationId = c.organization?._id || "-";

    const hasTaskResult = await TaskServices.findAll({ query: { workType: taskD.workType, caseId: taskD.caseId } });
    if (hasTaskResult.count > 0) {
        return res.status(500).send({ status: false, error: "Task already exists" });
    }

    const TaskResult = await TaskServices.create(taskD);
    if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

    return res.send(TaskResult);
    // return res.send({taskD, caseD});
});


router.get('/dev/requestInvoice', async (req, res) => {
    try {

        const CaseResult = await CaseServices.aggregate([
            {
                $match: { caseStatus: "DELIVERED", type: "CASE" }
            },
            {
                $lookup: {
                    from: "ninvoices",
                    localField: "_id",
                    foreignField: "cases._id",
                    as: "invoices"
                }
            },
            {
                $match: {
                    invoices: {
                        $size: 0
                    }
                }
            },
            {
                $project: {
                    caseNumber: 1,
                    caseStatus: 1,
                    type: 1,
                    ownerName: 1,
                    clientType: 1,
                    ownerName: 1,
                    clientType: 1,
                    typeOfReport: 1,
                    outward: 1,
                    propertyType: 1,
                    gstNo: 1,
                    client: 1,
                    branch: 1,
                }
            },
            {
                $limit: 50
            }
        ]
        );
        if (!CaseResult.status || CaseResult.error) return res.status(500).send(CaseResult);
        if (CaseResult.count <= 0) { return res.status(404).send({ status: false, error: "Case Not Found!" }) }
        const casesData = CaseResult.data.map((cs, id) => { return { ...cs, fees: 0 } });
        // const firstCase = casesData[0];

        const response = await Promise.all(casesData.map(async (firstCase, cid) => {
            const id = firstCase._id.toString();

            const organization = firstCase.organization;
            const gstNo = firstCase.gstNo || "";

            const invoice = {
                note: "SYSTEM GENERATE WHEN Lead/CASE CREATED",
                invoiceType: "PROFORMA",
                isAdvance: true,
                amount: 0,
                requestedBy: {
                    _id: "606456db5f122811d8d1d5bc",
                    firstName: "Administrator",
                    lastName: "",
                    profilePhoto: "users/606456db5f122811d8d1d5bc/profilePhoto_1684926242179_image.jpeg"
                },
                cases: firstCase,
                client: firstCase.client || {},
                branch: firstCase.branch || {},
                clientType: firstCase.clientType || {},
                organization: organization,
                gstNo: gstNo,
            };

            return await nInvoiceServices.create(invoice);
            // return await CaseServices.updateById(id, updateData)
        })).then(r => r)
        return res.json(response.length);
    }
    catch (err) {
        logger.error("nInvoice error request: " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
});

// const UserServices = require('../Services/CreateServices')('User');
// const coServices = require('../Services/CreateServices')('CreditOfficer');

// router.get('/dev/convert-cc', async (req, res) => {
//     try {
//         const creditOfficersRes = await UserServices.find({ query: { type: "CREDITOFFICER" }, limit: 10000 });
//         if (!creditOfficersRes.status) { return res.send(creditOfficersRes) }

//         const cdata = creditOfficersRes.data;

//         const response = await Promise.all(cdata.map(async (co, id) => {
//             console.log("co: ", co);
//             co = co.toJSON();
//             // delete co._id;
//             delete co.__v;
//             return await coServices.create({
//                 ...co,
//                 isDeleted: co.isdeleted,
//             })
//         })).then(r => r)
//         return res.send(response);
//     }
//     catch (err) {
//         logger.error("nInvoice error request: " + err.toString())
//         return res.status(500).send({ status: false, error: err });
//     }
// });

// router.get('/dev/update-case', async (req, res) => {
//     try {
//         const casesResult = await CaseServices.find({ query: { paymentStatus: "PENDING" }, limit: 500 });
//         if (!casesResult.status) { return res.send(casesResult) }

//         const cdata = casesResult.data;
//         const excelData = []; // Array to store data for Excel

//         const response = await Promise.all(cdata.map(async (co, cid) => {
//             co = co.toJSON();
//             const id = co._id.toString();

//             const hasInvoiceRes = await nInvoiceServices.find({ query: { 'cases._id': id, invoiceStatus: "COMPLETED" } });
//             if (!hasInvoiceRes.status || hasInvoiceRes.error) return;

//             if (hasInvoiceRes.count == 0) {
//                 console.log("=> NA ", co.caseNumber);
//                 await CaseServices.updateById(id, { paymentStatus: "NA" });
//                 excelData.push({
//                     caseNumber: co.caseNumber,
//                     pendingAmount: "N/A",
//                     receivedAmount: "N/A",
//                     newPaymentStatus: "NA",
//                     currentPaymentStatus: co.paymentStatus
//                 });
//                 return;
//             }

//             const pendingAmount = hasInvoiceRes.count != 0 ? hasInvoiceRes.data.map(invoice => invoice.pendingAmount).reduce((a, b) => a + b, 0) : -1;
//             const receivedAmount = hasInvoiceRes.count != 0 ? hasInvoiceRes.data.map(invoice => invoice.receivedAmount).reduce((a, b) => a + b, 0) : -1;

//             const hasPaymentRes = await PaymentServices.find({ query: { caseIds: { $in: id }, paymentStatus: "COMPLETED" } });
//             if (!hasPaymentRes.status || hasPaymentRes.error) return;

//             const newPaymentStatus = (pendingAmount == 0 || co.caseStatus === "COMPLETED") ? "RECEIVED" : (pendingAmount > 0 && receivedAmount != 0) ? "PART_RECEIVED" : "PENDING";
//             if (co.paymentStatus !== newPaymentStatus) {
//                 const updateData = {
//                     invoiceIds: hasInvoiceRes.data.map(invoice => invoice._id),
//                     paymentsIds: hasPaymentRes.data.map(payments => payments._id),
//                     paymentStatus: newPaymentStatus
//                 }
//                 console.log("=> updated obj", co.caseNumber, pendingAmount, receivedAmount, updateData);
//                 return Promise.resolve();
//                 // return await CaseServices.updateById(id, updateData)
//             } else {
//                 console.log("=> in else: cn: ", co.caseNumber, pendingAmount, receivedAmount, newPaymentStatus)
//                 return Promise.resolve();
//             }
//             // Add to Excel data array
//             excelData.push({
//                 caseNumber: co.caseNumber,
//                 pendingAmount: pendingAmount,
//                 receivedAmount: receivedAmount,
//                 newPaymentStatus: newPaymentStatus,
//                 currentPaymentStatus: co.paymentStatus
//             });
//         })).then(r => r);

//         // Generate Excel file using json2xls
//         const xls = json2xls(excelData);

//         // Define file path
//         const filePath = path.join(__dirname, 'cases_report.xlsx');

//         // Write file to disk
//         fs.writeFileSync(filePath, xls, 'binary');

//         // Send file as response
//         return res.download(filePath, 'cases_report.xlsx', (err) => {
//             if (err) throw err;

//             // Delete file after sending it
//             fs.unlinkSync(filePath);
//         });
//         return res.json(response.length);
//     }
//     catch (err) {
//         logger.error("nInvoice error request: " + err.toString())
//         return res.status(500).send({ status: false, error: err });
//     }
// });

router.get('/dev/invoice-payment-status', async (req, res) => {
    try {
        const invoiceResult = await nInvoiceServices.find({ query: { "cases.fees": { $eq: "0" }, "cases.pendingAmount": { $ne: "0" }, invoiceStatus: "COMPLETED", paymentStatus: "PENDING" }, limit: 100 });
        if (!invoiceResult.status) { return res.send(invoiceResult) }

        const cdata = invoiceResult.data;

        const response = await Promise.all(cdata.map(async (inv, cid) => {
            const id = inv._id.toString();

            inv.cases.map((cs, id) => {
                cs.fees = cs.pendingAmount || 0;
            })

            return nInvoiceServices.updateById(id, { cases: inv.cases });
        })).then(r => r)
        return res.json({ response, cdata });
    }
    catch (err) {
        logger.error("nInvoice /dev/invoice-payment-status request: " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
});


// without client name invoices
router.get('/dev/invoice-client-add', async (req, res) => {
    try {
        const invoiceResult = await nInvoiceServices.aggregate([
            {
                $match: {
                    client: { $exists: false },
                    invoiceStatus: "COMPLETED"
                }
            },
            {
                $lookup: {
                    from: "cases",
                    localField: "cases._id",
                    foreignField: "_id",
                    as: "caseDetails"
                }
            },
            {
                $addFields: { caseDetails: { $first: "$caseDetails" } }
            },
            {
                $addFields: {
                    client: "$caseDetails.client",
                    branch: "$caseDetails.branch"
                }
            },
            { $limit: 100 }
        ]);
        if (!invoiceResult.status) { return res.send(invoiceResult) }

        const cdata = invoiceResult.data;

        const response = await Promise.all(cdata.map(async (inv, cid) => {
            const id = inv._id.toString();

            return nInvoiceServices.updateById(id, { client: inv.client, branch: inv.branch });
        })).then(r => r)
        return res.json({ response, cdata });
    }
    catch (err) {
        logger.error("nInvoice /dev/invoice-client-add request: " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
});


const nInvoices = require('../Model/nInvoice');
const Payments = require('../Model/Payment');
const { ObjectId, ObjectID } = require("mongodb");
const Config = require("../Constants/Config");

router.get('/dev/payment-update', async (req, res) => {
    try {
        const paymentResults = await PaymentServices.find({
            query: {
                "paymentSettlement.cases.receivedAmount": 0,
                paymentStatus: "COMPLETED",
                "paymentSettlement.1": { $exists: false },
            },
            limit: 1
        });

        if (!paymentResults.status) { return res.send(paymentResults) }
        const paymentDatas = paymentResults.data;

        const response = await Promise.all(paymentDatas.map(async (pment, payid) => {
            const id = pment._id.toString();

            const relatedInvoices = await nInvoices.find({
                _id: { $in: pment.paymentSettlement.map((ps) => ObjectId(ps._id)) }
            });

            // Iterate over payment settlements to update receivedAmount
            for (const settlement of pment.paymentSettlement) {
                // Fetch related invoices for the settlement
                // Update receivedAmount for each case in the settlement
                settlement.cases.forEach((caseItem) => {
                    const relatedInvoice = relatedInvoices.find(
                        (invoice) => invoice.caseId === caseItem._id.toString()
                    );

                    if (relatedInvoice) {
                        // Update receivedAmount with invoice amount
                        caseItem.isCompleted = caseItem.isCompleted || true;
                        caseItem.invoicedAmount = +relatedInvoices.amount || 0;
                        caseItem.receivedAmount = +pment.voucher.voucher_amount || 0;
                        caseItem.pendingAmount = (relatedInvoice.amount || 0) - (caseItem.receivedAmount || 0);
                    }
                });
                settlement.isCompleted = true;
                settlement.receivedAmount = settlement.cases.reduce((acc, c) => acc + (c.receivedAmount || 0), 0);
                settlement.invoicedAmount = settlement.cases.reduce((acc, c) => acc + (c.invoicedAmount || 0), 0);
                settlement.pendingAmount = pment.voucher.voucher_amount || 0;
            }

            // Save the updated payment document
            await Payments.updateOne(
                { _id: ObjectId(id) },
                { $set: { paymentSettlement: pment.paymentSettlement } }
            );

            return { status: true, message: `Payment updated for ID: ${id}` };
        })
        );

        return res.json({ response });
    }
    catch (err) {
        logger.error("nInvoice /dev/invoice-payment-status request: " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
});

router.get('/dev/copy_new_case/:caseNumber', async (req, res) => {
    try {
        const { caseNumber } = req.params;
        const { token } = req.query;
        const existingCaseRes = await CaseServices.find({ query: { caseNumber } });

        if (token != moment().get('D')) { return res.status(501).send('Invalid Token'); }
        else { return res.status(501).send('success'); }

        if (!existingCaseRes.status || existingCaseRes.count == 0) {
            console.log(`No case found with caseNumber: ${caseNumber}`);
            return;
        }
        const existingCase = existingCaseRes.data[0].toJSON();
        const oldId = existingCase._id;
        const newId = new ObjectId();

        // Duplicate the existing document with a new _id
        const duplicatedCase = { ...existingCase, _id: newId };

        const updatedJsonObject = updateIdInJsonString(JSON.stringify(duplicatedCase), oldId, newId);

        // Update caseNumber and createdAt in the duplicated document
        updatedJsonObject._id = newId;
        updatedJsonObject.caseNumber = await generateNewCaseNumber();
        updatedJsonObject.createdAt = new Date();
        updatedJsonObject.createdBy = ObjectID(updatedJsonObject.createdBy);

        updatedJsonObject.outward = { status: false };
        updatedJsonObject.caseStatus = "CASE";
        updatedJsonObject.invoiceId = null;

        // Insert the duplicated document into the collection
        const newC = await CaseServices.create(updatedJsonObject);

        logger.error(`Case ${duplicatedCase.caseNumber} duplicated successfully with new caseNumber: ${JSON.stringify(updatedJsonObject)}`);
        const CDN = process.env.CDN_PATH;
        // Copy the entire folder
        await copyFolderRecursive(CDN + 'cases/' + oldId, CDN + 'cases/' + newId);
        return res.status(200).send({ status: true, newC })
    }
    catch (err) {
        logger.error("nInvoice /dev/copy new case request: " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
});


// Function to generate a new caseNumber
async function generateNewCaseNumber() {
    // Get the last created case from the collection
    const lastCaseRes = await CaseServices.find({ sort: { createdAt: -1 } });
    const lastCase = lastCaseRes.data[0] || { caseNumber: "000000" };

    var caseNo = lastCase.caseNumber;
    var caseNumber = parseInt(caseNo) + 1;
    const str = "" + caseNumber, pad = "000000"
    var answer = pad.substring(0, pad.length - str.length) + str;
    return answer;
}

function updateIdInJsonString(jsonString, oldId, newId) {
    try {
        const updatedJsonString = jsonString.replace(new RegExp(`\\b${oldId}\\b`, 'gi'), newId);
        const jsonObject = JSON.parse(updatedJsonString);
        return jsonObject;
    } catch (error) {
        console.error('Error parsing or updating JSON:', error);
        return null;
    }
}

function copyFolderRecursive(sourcePath, destinationPath) {
    // Ensure the destination directory exists
    if (!fs.existsSync(destinationPath)) {
        fs.mkdirSync(destinationPath);
    }

    // Read the contents of the source directory
    const files = fs.readdirSync(sourcePath);
    console.log("->OLD: ", sourcePath, "\n New: ", destinationPath);

    files.forEach((file) => {
        const sourceFilePath = path.join(sourcePath, file);
        const destinationFilePath = path.join(destinationPath, file);
        // Check if the current item is a directory
        const isDirectory = fs.statSync(sourceFilePath).isDirectory();

        if (isDirectory) {
            // Recursively copy subdirectories
            copyFolderRecursive(sourceFilePath, destinationFilePath);
        } else {
            // Copy individual files
            fs.copyFileSync(sourceFilePath, destinationFilePath);
        }
    });
}


module.exports = router;